﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Linq.Expressions;

namespace ParsecParser
{
    public static class CombatEventFilters
    {
        public static Expression<Func<CombatLogEvent, bool>> IsSource()
        {
            return PredicateBuilder.False<CombatLogEvent>().Or(x=>x.Source.IsThisPlayer);
        }
        public static Expression<Func<CombatLogEvent, bool>> IsDamage()
        {
            return PredicateBuilder.False<CombatLogEvent>().Or(x => x.EffectName.EntityId == CombatLogConstants.EffectNames.Damage);
        }
        public static Expression<Func<CombatLogEvent, bool>> IsTarget()
        {
            return PredicateBuilder.False<CombatLogEvent>().Or(x => x.Target.IsThisPlayer);
        }
        public static Expression<Func<CombatLogEvent, bool>> IsNotTarget()
        {
            return PredicateBuilder.False<CombatLogEvent>().Or(x => !x.Target.IsThisPlayer);
        }
        public static Expression<Func<CombatLogEvent, bool>> IsNotSource()
        {
            return PredicateBuilder.False<CombatLogEvent>().Or(x => !x.Source.IsThisPlayer);
        }
        public static Expression<Func<CombatLogEvent, bool>> IsHeal()
        {
            return PredicateBuilder.False<CombatLogEvent>().Or(x => x.EffectName.EntityId == CombatLogConstants.EffectNames.Heal);
        }
        public static Expression<Func<CombatLogEvent, bool>> IsEffect()
        {
            return PredicateBuilder.False<CombatLogEvent>().Or(x => x.EffectType.EntityId == CombatLogConstants.EffectTypes.ApplyEffect && x.EffectName.EntityId != CombatLogConstants.EffectNames.Heal && x.EffectName.EntityId != CombatLogConstants.EffectNames.Damage).Or(x=>x.EffectType.EntityId == CombatLogConstants.EffectTypes.RemoveEffect);
        }
        public static Expression<Func<CombatLogEvent, bool>> TargetIsNotNpc()
        {
            return PredicateBuilder.False<CombatLogEvent>().Or(x => x.Target.IsPlayer || x.Target.IsPlayerCompanion);
        }
        public static Expression<Func<CombatLogEvent, bool>> IsAbility()
        {
            long[] grouping = {CombatLogConstants.EffectNames.AbilityActivate, CombatLogConstants.EffectNames.AbilityCancel, CombatLogConstants.EffectNames.AbilityDeactivate, CombatLogConstants.EffectNames.AbilityInterrupt};
            return PredicateBuilder.False<CombatLogEvent>().Or(x => x.EffectType.EntityId == CombatLogConstants.EffectTypes.Event && grouping.Contains(x.EffectName.EntityId));
        }
        public static Expression<Func<CombatLogEvent, bool>> IsEvent()
        {
            long[] grouping = { CombatLogConstants.EffectNames.AbilityActivate, CombatLogConstants.EffectNames.AbilityCancel, CombatLogConstants.EffectNames.AbilityDeactivate, CombatLogConstants.EffectNames.AbilityInterrupt };
            return PredicateBuilder.False<CombatLogEvent>().Or(x => x.EffectType.EntityId == CombatLogConstants.EffectTypes.Event && !grouping.Contains(x.EffectName.EntityId));
        }
        public static Expression<Func<CombatLogEvent, bool>> IsOutgoingDamage()
        {
            return IsSource().And(IsDamage());
        }
        public static Expression<Func<CombatLogEvent, bool>> IsIncomingDamage()
        {
            return IsTarget().And(IsDamage());
        }
        public static Expression<Func<CombatLogEvent, bool>> IsOutgoingHeal()
        {
            return IsSource().And(TargetIsNotNpc()).And(IsHeal());
        }
        public static Expression<Func<CombatLogEvent, bool>> IsIncomingHeal()
        {
            return IsTarget().And(IsHeal());
        }
        public static Expression<Func<CombatLogEvent, bool>> IsOthersEffect()
        {
            return IsNotTarget().And(IsEffect());
        }
        public static Expression<Func<CombatLogEvent, bool>> IsCharacterEffect()
        {
            return IsTarget().And(IsEffect());
        }
        public static IEnumerable<CombatLogEvent> ApplyFilter(this IEnumerable<CombatLogEvent> source, CombatLogFilterParameters p)
        {
            if (p.ShowAll || p.EffectiveShowAll)
                return source;
            else
            {
                var predicate = PredicateBuilder.False<CombatLogEvent>();
                if (p.ShowAbilities)
                    predicate = predicate.Or(IsAbility());
                if (p.ShowIncomingDamage)
                    predicate = predicate.Or(IsIncomingDamage());
                if (p.ShowOthersEffects)
                    predicate = predicate.Or(IsOthersEffect());
                if (p.ShowIncomingHeals)
                    predicate = predicate.Or(IsIncomingHeal());
                if (p.ShowOutgoingHeals)
                    predicate = predicate.Or(IsOutgoingHeal());
                if (p.ShowOutgoingDamage)
                    predicate = predicate.Or(IsOutgoingDamage());
                if (p.ShowEvents)
                    predicate = predicate.Or(IsEvent());
                if (p.ShowCharacterEffects)
                    predicate = predicate.Or(IsCharacterEffect());
                return source.AsQueryable().Where(predicate);
            }
        }
        public static IEnumerable<string> ApplyFilter(this IEnumerable<string> source, CombatLogFilterParameters p)
        {
            if (!string.IsNullOrEmpty(p.KeyWordFilter))
                return source.Where(x => x.ContainsString(p.KeyWordFilter, true));
            return source;
        }
    }
}
