﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace ParsecParser
{
    public class RealTimeLogReader : IDisposable
    {
        private System.Timers.Timer PollTimer;
        private FileStream FileStream;
        private StreamReader Reader;
        private long previousPosition;
        private bool IsWatchingFile;
        private bool IsWatchingFolder;
        private bool IsPolling;
        public bool TickWhenLogIsNotUpdated { get; set; }
        public string FileName { get; set; }
        public string FolderPath { get; set; }
        public long PollMilliseconds {get;set;}
        public event EventHandler<OnLogUpdatedEventArgs> LogUpdated;
        public event EventHandler<OnNewLogCreatedEventArgs> NewLogCreated;
        public event EventHandler<ErrorEventArgs> Error;
        public RealTimeLogReader(string filename, string folderPath, EventHandler<OnLogUpdatedEventArgs> onLogUpdated, EventHandler<OnNewLogCreatedEventArgs> onNewLogCreated, EventHandler<ErrorEventArgs> onError)
        {
            FileName = filename;
            FolderPath = folderPath;
            previousPosition = 0;
            PollMilliseconds = 3000;
            if (onLogUpdated != null)
                LogUpdated += onLogUpdated;
            if (onNewLogCreated != null)
                NewLogCreated += onNewLogCreated;
            if (onError != null)
                Error += onError;
        }
        public List<string> ReadNewContent()
        {
            if (string.IsNullOrEmpty(FileName))
                throw new ArgumentException("FileName cannot be null");
            if (FileStream != null && String.Compare(FileStream.Name, FileName, true) != 0)
            {
                StopReading();
            }
            if (FileStream == null)
            {
                if (!File.Exists(FileName))
                    throw new Exception("File Not Found");
                FileStream = new FileStream(FileName, FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
                Reader = null;
                if (previousPosition > 0)
                {
                    FileStream.Seek(previousPosition, SeekOrigin.Begin);
                }
            }
            if (Reader == null)
            {
                Reader = new StreamReader(FileStream, Encoding.GetEncoding("ISO-8859-15"));
            }
            if (previousPosition > 0)
            {
                Reader.DiscardBufferedData();
            }
            if (Reader.BaseStream.Position > Reader.BaseStream.Length)
            {
                Reader.BaseStream.Seek(Reader.BaseStream.Length, SeekOrigin.Begin);
            }
            List<string> rval = new List<string>();
            while (!Reader.EndOfStream)
            {
                string newLine = Reader.ReadLine();
                if (newLine != null)
                    rval.Add(newLine);
                else
                    throw new Exception("Line is null");
            }
            previousPosition = Reader.BaseStream.Position;
            return rval;
        }
        public void StartWatchingFolder()
        {
            IsWatchingFolder = true;
            if (string.IsNullOrEmpty(FolderPath))
                throw new ArgumentException("FolderPath cannot be null");
            if (!Directory.Exists(FolderPath))
                throw new ArgumentException("Folder does not exist");
            if (PollTimer == null || !IsWatchingFile)
            {
                StartPolling();
            }
        }
        public void StartWatchingFile()
        {
            IsWatchingFile = true;
            if (string.IsNullOrEmpty(FileName))
                throw new ArgumentException("FileName cannot be null");
            if (!File.Exists(FileName))
                throw new ArgumentException("File does not exist");
            if (PollTimer == null || !IsWatchingFolder)
            {
                StartPolling();
            }
        }
        public void StopWatchingFile()
        {
            IsWatchingFile = false;
            if (PollTimer != null && !IsWatchingFolder)
            {
                StopPolling();
            }
        }
        public void StopWatchingFolder()
        {
            IsWatchingFolder = false;
            if (PollTimer != null && !IsWatchingFile)
            {
                StopPolling();
            }
        }
        private void StopPolling()
        {
            if (IsPolling)
            {
                PollTimer.AutoReset = false;
                PollTimer.Elapsed -= PollTimer_Elapsed;
                PollTimer.Stop();
                IsPolling = false;
                PollTimer.Dispose();
            }
        }
        private void StartPolling()
        {
            if (!IsPolling)
            {
                PollTimer = new System.Timers.Timer(PollMilliseconds);
                PollTimer.AutoReset = true;
                PollTimer.Elapsed += new System.Timers.ElapsedEventHandler(PollTimer_Elapsed);
                PollTimer.Start();
                IsPolling = true;
            }
        }
        public void StopReading()
        {
            if (Reader != null)
            {
                Reader.Close();
                Reader = null;
            }
            else if (FileStream != null)
                FileStream.Close();
            FileStream = null;
            previousPosition = 0;
        }
        private void PollTimer_Elapsed(object source, System.Timers.ElapsedEventArgs e)
        {
            try
            {
                if (IsWatchingFile)
                {
                    List<string> newLogLines = ReadNewContent();
                    if (newLogLines.Count > 0 || TickWhenLogIsNotUpdated)
                    {
                        OnLogUpdated(new OnLogUpdatedEventArgs() { NewLines = newLogLines });
                    }
                }
                if (IsWatchingFolder)
                {
                    string newFile = GetNewestFile(FolderPath);
                    if (newFile != FileName)
                    {
                        OnNewLogCreated(new OnNewLogCreatedEventArgs() { FileName = newFile });
                    }
                }
                if (!IsWatchingFile && !IsWatchingFolder)
                {
                    IsPolling = true;
                    StopPolling();
                }
            }
            catch (Exception ex)
            {
                if (!OnError(new ErrorEventArgs(ex)))
                {
                    throw;
                }
            }
        }
        protected virtual void OnLogUpdated(OnLogUpdatedEventArgs e)
        {
            EventHandler<OnLogUpdatedEventArgs> handler = LogUpdated;
            if (handler != null)
            {
                handler(this, e);
            }
            else if (IsWatchingFile)
            {
                StopWatchingFile();
                StopWatchingFolder();
                StopReading();
            }
        }
        protected virtual void OnNewLogCreated(OnNewLogCreatedEventArgs e)
        {
            EventHandler<OnNewLogCreatedEventArgs> handler = NewLogCreated;
            if (handler != null)
            {
                handler(this, e);
            }
            else if (IsWatchingFolder)
            {
                StopWatchingFile();
                StopWatchingFolder();
                StopReading();
            }
        }
        protected virtual bool OnError(ErrorEventArgs e)
        {
            EventHandler<ErrorEventArgs> handler = Error;
            if (handler != null)
            {
                handler(this, e);
                return true;
            }
            else
                return false;
        }

        public void Dispose()
        {
            StopWatchingFile();
            StopWatchingFolder();
            StopReading();
        }

        public static string GetNewestFile(string folderPath)
        {
            var d = new DirectoryInfo(folderPath);
            var newFile = d.GetFiles().OrderByDescending(x => x.LastWriteTime).FirstOrDefault();
            if (newFile != null)
                return newFile.FullName;
            else
                return null;
        }
    }
    public class OnLogUpdatedEventArgs : EventArgs
    {
        public List<string> NewLines { get; set; }
    }
    public class OnNewLogCreatedEventArgs : EventArgs
    {
        public string FileName { get; set; }
    }
}
